@@ -39,25 +39,12 @@
 #include <time.h>
 #include <errno.h>
 #include <ctype.h>
-
-#ifdef HAVE_DIRENT_H
-#  include <dirent.h>
-#else /* not HAVE_DIRENT_H */
-#  define dirent direct
-#  ifdef HAVE_SYS_NDIR_H
-#    include <sys/ndir.h>
-#  endif /* HAVE_SYS_NDIR_H */
-#  ifdef HAVE_SYS_DIR_H
-#    include <sys/dir.h>
-#  endif /* HAVE_SYS_DIR_H */
-#  ifdef HAVE_NDIR_H
-#    include <ndir.h>
-#  endif /* HAVE_NDIR_H */
-#endif /* HAVE_DIRENT_H  */
-
+#include <dirent.h>
 #include <unistd.h>
 
 #include "dirname.h"
+#include "stat-time.h"
+#include "timespec.h"
 #include "xvasprintf.h"
 
 #include "gettext.h"
@@ -67,6 +54,7 @@
 
 #include "error.h"
 #include "hashtable.h"
+#include "orderfiles.h"
 #include "security.h"
 
 #include "mydbm.h"
@@ -156,7 +144,7 @@ void test_manfile (const char *file, const char *path)
 
 	/* to get mtime info */
 	(void) lstat (file, &buf);
-	info._st_mtime = buf.st_mtime;
+	info.mtime = get_stat_mtime (&buf);
 
 	/* check that our file actually contains some data */
 	if (buf.st_size == 0) {
@@ -177,8 +165,8 @@ void test_manfile (const char *file, const char *path)
 	 */
 	if (exists) {
 		if (strcmp (exists->comp, info.comp ? info.comp : "-") == 0) {
-			if (exists->_st_mtime == info._st_mtime 
-			    && exists->id < WHATIS_MAN) {
+			if (timespec_cmp (exists->mtime, info.mtime) == 0 &&
+			    exists->id < WHATIS_MAN) {
 				free_mandata_struct (exists);
 				free (manpage);
 				return;
@@ -323,6 +311,8 @@ static inline void add_dir_entries (const char *path, char *infile)
 	int len;
 	struct dirent *newdir;
 	DIR *dir;
+	char **names;
+	size_t names_len, names_max, i;
 
 	manpage = xasprintf ("%s/%s/", path, infile);
 	len = strlen (manpage);
@@ -338,19 +328,36 @@ static inline void add_dir_entries (const char *path, char *infile)
 		free (manpage);
                 return;
         }
-        
+
+	names_len = 0;
+	names_max = 1024;
+	names = XNMALLOC (names_max, char *);
+
         /* strlen(newdir->d_name) could be replaced by newdir->d_reclen */
-        
-	while ( (newdir = readdir (dir)) )
-		if (!(*newdir->d_name == '.' && 
-		      strlen (newdir->d_name) < (size_t) 3)) {
-			manpage = appendstr (manpage, newdir->d_name, NULL);
-			test_manfile (manpage, path);
-			*(manpage + len) = '\0';
+
+	while ((newdir = readdir (dir)) != NULL) {
+		if (*newdir->d_name == '.' &&
+		    strlen (newdir->d_name) < (size_t) 3)
+			continue;
+		if (names_len >= names_max) {
+			names_max *= 2;
+			names = xnrealloc (names, names_max, sizeof (char *));
 		}
-		
-	free (manpage);
+		names[names_len++] = xstrdup (newdir->d_name);
+	}
 	closedir (dir);
+
+	order_files (infile, names, names_len);
+
+	for (i = 0; i < names_len; ++i) {
+		manpage = appendstr (manpage, names[i], NULL);
+		test_manfile (manpage, path);
+		*(manpage + len) = '\0';
+		free (names[i]);
+	}
+
+	free (names);
+	free (manpage);
 }
 
 #ifdef SECURE_MAN_UID
@@ -425,12 +432,11 @@ static void mkcatdirs (const char *mandir, const char *catdir)
  * any dirs of the tree that have been modified (ie added to) will then be
  * scanned for new files, which are then added to the db.
  */
-static int testmandirs (const char *path, const char *catpath, time_t last,
-			int create)
+static int testmandirs (const char *path, const char *catpath,
+			struct timespec last, int create)
 {
 	DIR *dir;
 	struct dirent *mandir;
-	struct stat stbuf;
 	int amount = 0;
 	int created = 0;
 
@@ -449,6 +455,9 @@ static int testmandirs (const char *path, const char *catpath, time_t last,
 	}
 
 	while( (mandir = readdir (dir)) ) {
+		struct stat stbuf;
+		struct timespec mtime;
+
 		if (strncmp (mandir->d_name, "man", 3) != 0)
 			continue;
 
@@ -458,11 +467,14 @@ static int testmandirs (const char *path, const char *catpath, time_t last,
 			continue;
 		if (!S_ISDIR(stbuf.st_mode))		/* not a directory */
 			continue;
-		if (last && stbuf.st_mtime <= last) {
+		mtime = get_stat_mtime (&stbuf);
+		if (last.tv_sec && timespec_cmp (mtime, last) <= 0) {
 			/* scanned already */
-			debug ("%s modified %ld, db modified %ld\n",
-			       mandir->d_name, (long) stbuf.st_mtime,
-			       (long) last);
+			debug ("%s modified %ld.%09ld, "
+			       "db modified %ld.%09ld\n",
+			       mandir->d_name,
+			       (long) mtime.tv_sec, mtime.tv_nsec,
+			       (long) last.tv_sec, last.tv_nsec);
 			continue;
 		}
 
@@ -525,25 +537,12 @@ static int testmandirs (const char *path, const char *catpath, time_t last,
 	return amount;
 }
 
-/* update the time key stored within `database' */
-void update_db_time (void)
+/* update the modification timestamp of `database' */
+static void update_db_time (void)
 {
-	datum key, content;
-#ifdef FAST_BTREE
-	datum key1, content1;
-#endif /* FAST_BTREE */
+	struct timespec now;
 
-	memset (&key, 0, sizeof key);
-	memset (&content, 0, sizeof content);
-#ifdef FAST_BTREE
-	memset (&key1, 0, sizeof key);
-	memset (&content1, 0, sizeof content);
-#endif
-
-	MYDBM_SET (key, xstrdup (KEY));
-	MYDBM_SET (content, xasprintf ("%ld", (long) time (NULL)));
-
-	/* Open the db in RW to store the $mtime$ ID */
+	/* Open the db in RW to update its mtime */
 	/* we know that this should succeed because we just updated the db! */
 	dbf = MYDBM_RWOPEN (database);
 	if (dbf == NULL) {
@@ -561,58 +560,26 @@ void update_db_time (void)
 				       _("can't update index cache %s"),
 				       database);
 		}
-		free (MYDBM_DPTR (content));
 		return;
 	}
-#ifndef FAST_BTREE
-	MYDBM_REPLACE (dbf, key, content);
-#else /* FAST_BTREE */
-	MYDBM_SET (key1, KEY);
-
-	(dbf->seq) (dbf, (DBT *) &key1, (DBT *) &content1, R_CURSOR);
-	
-	if (strcmp (MYDBM_DPTR (key1), MYDBM_DPTR (key)) == 0)
-		(dbf->put) (dbf, (DBT *) &key, (DBT *) &content, R_CURSOR);
-	else
-		(dbf->put) (dbf, (DBT *) &key, (DBT *) &content, 0);
-#endif /* !FAST_BTREE */
+	now.tv_sec = 0;
+	now.tv_nsec = UTIME_NOW;
+	MYDBM_SET_TIME (dbf, now);
 
 	MYDBM_CLOSE (dbf);
-	free (MYDBM_DPTR (key));
-	free (MYDBM_DPTR (content));
-}
-
-/* remove the db's time key - called prior to update_db if we want
-   to `force' a full consistency check */
-void reset_db_time (void)
-{
-	datum key;
-
-	memset (&key, 0, sizeof key);
-
-	MYDBM_SET (key, xstrdup (KEY));
-
-	/* we don't really care if we can't open it RW - it's not fatal */
-	dbf = MYDBM_RWOPEN (database);
-	if (dbf == NULL) {
-		debug_error ("reset_db_time(): can't open db");
-		return;
-	}
-
-	MYDBM_DELETE (dbf, key);
-	debug ("reset_db_time()\n");
-	MYDBM_CLOSE (dbf);
-	free (MYDBM_DPTR (key));
 }
 
 /* routine to prepare/create the db prior to calling testmandirs() */
 int create_db (const char *manpath, const char *catpath)
 {
+	struct timespec time_zero;
 	int amount;
-	
+
 	debug ("create_db(%s): %s\n", manpath, database);
 
-	amount = testmandirs (manpath, catpath, (time_t) 0, 1);
+	time_zero.tv_sec = 0;
+	time_zero.tv_nsec = 0;
+	amount = testmandirs (manpath, catpath, time_zero, 1);
 
 	if (amount) {
 		update_db_time ();
@@ -624,7 +591,7 @@ int create_db (const char *manpath, const char *catpath)
 }
 
 /* Make sure an existing database is essentially sane. */
-int sanity_check_db (void)
+static int sanity_check_db (void)
 {
 	datum key;
 
@@ -661,26 +628,12 @@ int update_db (const char *manpath, const char *catpath)
 		dbf = NULL;
 	}
 	if (dbf) {
-		datum key, content;
+		struct timespec mtime = MYDBM_GET_TIME (dbf);
 		int new;
 
-		memset (&key, 0, sizeof key);
-		memset (&content, 0, sizeof content);
-
-		MYDBM_SET (key, xstrdup (KEY));
-		content = MYDBM_FETCH (dbf, key);
-		MYDBM_CLOSE (dbf);
-		free (MYDBM_DPTR (key));
-
-		debug ("update_db(): %ld\n",
-		       MYDBM_DPTR (content) ? atol (MYDBM_DPTR (content)) : 0L);
-		if (MYDBM_DPTR (content)) {
-			new = testmandirs (
-				manpath, catpath,
-				(time_t) atol (MYDBM_DPTR (content)), 0);
-			MYDBM_FREE (MYDBM_DPTR (content));
-		} else
-			new = testmandirs (manpath, catpath, (time_t) 0, 0);
+		debug ("update_db(): %ld.%09ld\n",
+		       (long) mtime.tv_sec, mtime.tv_nsec);
+		new = testmandirs (manpath, catpath, mtime, 0);
 
 		if (new) {
 			update_db_time ();
@@ -757,7 +710,7 @@ pointers_next:
  * out that this is better handled in look_for_file() itself.
  */
 static int count_glob_matches (const char *name, const char *ext,
-			       char **source, long db_mtime)
+			       char **source, struct timespec db_mtime)
 {
 	char **walk;
 	int count = 0;
@@ -774,7 +727,8 @@ static int count_glob_matches (const char *name, const char *ext,
 			       "because stat failed\n", *walk);
 			continue;
 		}
-		if (db_mtime != -1 && statbuf.st_mtime <= db_mtime) {
+		if (db_mtime.tv_sec != (time_t) -1 &&
+		    timespec_cmp (get_stat_mtime (&statbuf), db_mtime) <= 0) {
 			debug ("count_glob_matches: excluding %s, "
 			       "no newer than database\n", *walk);
 			continue;
@@ -799,10 +753,14 @@ static int count_glob_matches (const char *name, const char *ext,
 static int purge_normal (const char *name, struct mandata *info,
 			 char **found)
 {
+	struct timespec t;
+
 	/* TODO: On some systems, the cat page extension differs from the
 	 * man page extension, so this may be too strict.
 	 */
-	if (count_glob_matches (name, info->ext, found, -1))
+	t.tv_sec = -1;
+	t.tv_nsec = -1;
+	if (count_glob_matches (name, info->ext, found, t))
 		return 0;
 
 	if (!opt_test)
@@ -816,7 +774,8 @@ static int purge_normal (const char *name, struct mandata *info,
 
 /* Decide whether to purge a reference to a WHATIS_MAN or WHATIS_CAT page. */
 static int purge_whatis (const char *path, int cat, const char *name,
-			 struct mandata *info, char **found, long db_mtime)
+			 struct mandata *info, char **found,
+			 struct timespec db_mtime)
 {
 	/* TODO: On some systems, the cat page extension differs from the
 	 * man page extension, so this may be too strict.
@@ -857,13 +816,17 @@ static int purge_whatis (const char *path, int cat, const char *name,
 		/* Does the real page still exist? */
 		char **real_found;
 		int save_debug = debug_level;
+		struct timespec t;
+
 		debug_level = 0;
 		real_found = look_for_file (path, info->ext,
 					    info->pointer, cat, LFF_MATCHCASE);
 		debug_level = save_debug;
 
+		t.tv_sec = -1;
+		t.tv_nsec = -1;
 		if (count_glob_matches (info->pointer, info->ext, real_found,
-					-1))
+					t))
 			return 0;
 
 		if (!opt_test)
@@ -917,7 +880,8 @@ static int check_multi_key (const char *name, const char *content)
 /* Go through the database and purge references to man pages that no longer
  * exist.
  */
-int purge_missing (const char *manpath, const char *catpath)
+int purge_missing (const char *manpath, const char *catpath,
+		   int will_run_mandb)
 {
 #ifdef NDBM
 	char *dirfile;
@@ -926,7 +890,7 @@ int purge_missing (const char *manpath, const char *catpath)
 	int db_exists;
 	datum key;
 	int count = 0;
-	long db_mtime = -1;
+	struct timespec db_mtime;
 
 #ifdef NDBM
 	dirfile = xasprintf ("%s.dir", database);
@@ -947,29 +911,11 @@ int purge_missing (const char *manpath, const char *catpath)
 		gripe_rwopen_failed ();
 		return 0;
 	}
-
-	/* Extract the database mtime. */
-	key = MYDBM_FIRSTKEY (dbf);
-	while (MYDBM_DPTR (key) != NULL) {
-		datum content, nextkey;
-
-		if (STREQ (MYDBM_DPTR (key), KEY)) {
-			content = MYDBM_FETCH (dbf, key);
-			if (MYDBM_DPTR (content)) {
-				errno = 0;
-				db_mtime = strtol (MYDBM_DPTR (content), NULL,
-						   10);
-				if (errno)
-					db_mtime = -1;
-				MYDBM_FREE (MYDBM_DPTR (key));
-				break;
-			}
-		}
-
-		nextkey = MYDBM_NEXTKEY (dbf, key);
-		MYDBM_FREE (MYDBM_DPTR (key));
-		key = nextkey;
+	if (!sanity_check_db ()) {
+		MYDBM_CLOSE (dbf);
+		return 0;
 	}
+	db_mtime = MYDBM_GET_TIME (dbf);
 
 	key = MYDBM_FIRSTKEY (dbf);
 
@@ -1052,6 +998,12 @@ int purge_missing (const char *manpath, const char *catpath)
 	}
 
 	MYDBM_REORG (dbf);
+	if (will_run_mandb)
+		/* Reset mtime to avoid confusing mandb into not running.
+		 * TODO: It would be better to avoid this by only opening
+		 * the database once between here and mandb.
+		 */
+		MYDBM_SET_TIME (dbf, db_mtime);
 	MYDBM_CLOSE (dbf);
 	return count;
 }
